"""
This event does two things.

First, it assigns a Fleets Pools and Groups upon initialization.

Secondly, it determines if we should perform auto shutdown of the Worker. 
Auto Shutdown is essentially terminating the instance if the Worker on the
instance has been idle for too long.
"""


###############################################################
## Imports
###############################################################
from __future__ import absolute_import
import boto3
import datetime
import dateutil
import dateutil.parser
import json
import os
import six
import sys
import time
from io import open
import traceback

from Deadline.Dash import DashFleetInstance, DashFleetRequest
from Deadline.Events import DeadlineEventListener
from Deadline.Scripting import ClientUtils, RepositoryUtils
from Deadline.Slaves import SlaveInfo, SlaveSettings

# Add the events folder to the PYTHONPATH.
eventPath = RepositoryUtils.GetEventPluginDirectory("AWSPortal")

if not eventPath in sys.path:
    sys.path.append( eventPath )

from AWSPortalUtils import PoolGroupCheck

from typing import Dict, List, Optional, Text, Tuple, Union

class AWSPortalEventError(Exception):
    pass

##############################################################################################
## This is the function called by Deadline to get an instance of the AWS Portal event listener.
##############################################################################################
def GetDeadlineEventListener():
    # type: () -> AWSPortalEventListener
    return AWSPortalEventListener()

def CleanupDeadlineEventListener(eventListener):
    # type: (DeadlineEventListener) -> None
    eventListener.Cleanup()


###############################################################
## The AWSPortal event listener class.
###############################################################
class AWSPortalEventListener(DeadlineEventListener):
    def __init__( self ):
        # type: () -> None
        if sys.version_info.major == 3:
            super().__init__()
        self.OnSlaveInfoUpdatedCallback += self.OnSlaveInfoUpdated
        self.OnSlaveStoppedCallback += self.OnSlaveStopped
        self.OnSlaveStartedCallback += self.OnSlaveStarted
        self.LastRenderingTimePrefix = "Last Rendering Time: "
        self.Disabled = "Disabled"
        self.Rendering = "Rendering"
        self.WorkersStatusFolderPath = os.path.join(Py_GetCurrentUserHomeDirectory(), "AWSPortalWorkerStatus")
        if not os.path.isdir(self.WorkersStatusFolderPath):
            os.makedirs(self.WorkersStatusFolderPath)

    def Cleanup( self ):
        # type: () -> None
        del self.OnSlaveInfoUpdatedCallback
        del self.OnSlaveStoppedCallback
        del self.OnSlaveStartedCallback

    def DeserializeGroupsAndPools(self, groupsAndPools):
        # type: (str) -> Tuple[List[str], List[str]]
        """
        Take in a json string that contains lists of groups and pools, change it to
        a dictionary and return the lists of groups and pools.

        :param groupsAndPools:
        :return:
        """

        groupsAndPoolsDict = json.loads( groupsAndPools )

        return groupsAndPoolsDict['groups'], groupsAndPoolsDict['pools']

    def SetGroupsAndPools(self, slaveName, groups, pools):
        # type: (str, List[str], List[str]) -> None
        """
        Set the Groups and pools for the given Worker
        """

        slaveSettings = Py_GetSlaveSettings( slaveName, True )

        slaveSettings.Groups = groups
        slaveSettings.Pools = pools

        Py_SaveSlaveSettings( slaveSettings )

        #if Pools and Groups have been set, we no longer need to check them. Set the last check time to Max.
        self.WritePoolGroupCheckTime(datetime.datetime.max)

    def OnSlaveStarted( self, slaveName ):
        # type: (str) -> None
        self.LogInfo( "On Worker Started" )
        slaveInfo = Py_GetSlaveInfo( slaveName, True )

        if hasattr(slaveInfo, "AWSInfo"):
            if "AWSInstanceId" in slaveInfo.AWSInfo.Keys:
                self.LogInfo("This instance is running on AWS - AWSInstanceId is set:" + str(slaveInfo.AWSInfo["AWSInstanceId"]))#jusbla
                instanceId = slaveInfo.AWSInfo["AWSInstanceId"]
                self.AssignPoolsAndGroups( slaveName, instanceId )

    def GetStatus(self, workerName):
        # type: (str) -> Optional[str]
        """
        Return the status of the Worker. If the status of the Worker cannot
        be determined, return None.

        The status would look one of the three examples below
            Disabled
            Rendering
            Last Rendering Time: 2020-03-09 21:37:56.753000
        """
        workerStatusFilePath = os.path.join(self.WorkersStatusFolderPath, workerName)

        if not os.path.isfile(workerStatusFilePath):
            self.LogInfo("GetStatus: Status file not found")#jusbla
            return None

        with open(workerStatusFilePath, "r") as f:
            status = f.readline().rstrip()
            if status:
                self.LogInfo("Worker status is " + status)#jusbla
                return status
        
        return None

    def UpdateStatus(self, workerName, status):
        # type: (str, Union[str, Text]) -> None
        """
        Update the status of the Worker.
        """
        status = six.ensure_text(status)

        self.LogInfo(f"Writing '{status}' to Worker status file in {os.path.join(self.WorkersStatusFolderPath, workerName)}")

        with open(os.path.join(self.WorkersStatusFolderPath, workerName), "w") as f:
            f.write(status)

    def GetLastRenderingTime(self, status):
        # type: (str) -> datetime.datetime
        """
        Get the last rendering time from the Status.
        """
        if not status.startswith(self.LastRenderingTimePrefix):
            raise AWSPortalEventError("Unable to parse status")
        return dateutil.parser.parse(status.replace(self.LastRenderingTimePrefix, ''))

    def DeleteStatus(self, workerName):
        # type: (str) -> None
        """
        Delete the status of the Worker.
        """
        os.remove(os.path.join(self.WorkersStatusFolderPath, workerName))

    def OnSlaveInfoUpdated(self, slaveName, slaveInfo):
        # type: (str, SlaveInfo) -> None
        """
        If we have not determined the status of the Worker yet, then we check if
        Auto Shutdown is enabled on the Worker.
        If Auto Shutdown is disabled, we set the status to Disabled, and do not perform
        Auto Shutdown.
        If Auto Shutdown is enabled, we set the status to Rendering or the current time.
        - if the status is rendering, we should not Shutdown the instance.
        - if the status is the timestamp, we check how long it has been idle for, and if it
          has been idle for too long we perform a Shutdown sequence.
        """
        try:
            verbose = self.GetBooleanConfigEntryWithDefault("Verbose", False)

            if verbose:
                self.LogInfo( "AWS Portal Plugin - On Worker Info Updated Started" )

            if hasattr(slaveInfo, "AWSInfo"):
                if "AWSInstanceId" in slaveInfo.AWSInfo.Keys:
                    instanceId = slaveInfo.AWSInfo["AWSInstanceId"]
                    self.LogInfo("This instance is running on AWS - AWSInstanceId is set:" + str(slaveInfo.AWSInfo["AWSInstanceId"]))#jusbla
                    now = datetime.datetime.utcnow()

                    if now - PoolGroupCheck.GetLastPoolGroupCheckinTime() > datetime.timedelta(minutes=2):
                        # Check to see if Pools and Groups have been set yet.
                        slaveSettings = Py_GetSlaveSettings( slaveName, False )
                        if len( slaveSettings.SlaveGroups ) == 0 or len( slaveSettings.SlavePools ) == 0:
                            self.AssignPoolsAndGroups( slaveName, instanceId )
                        else:
                            #if the group and pool have been set to a value, we should not check anymore, set the check time to Max.
                            self.WritePoolGroupCheckTime( datetime.datetime.max )
                                      
                    if verbose:
                        self.LogInfo("Checking if the Worker is rendering and if it should automatically shut down")

                    status = self.GetStatus(slaveName) # this pulls the status from the file written by self.UpdateStatus
                    shutdownSetting = int(self.CheckAutoShutdown( slaveInfo, verbose ))
                    if verbose:
                        self.LogInfo(f"Shutdown setting is {shutdownSetting}")
                        self.LogInfo(f"Autoshutdown is {Py_GetAWSPortalRequestTypeEnumValue('AutoShutdown')}")
                        self.LogInfo(f"Maintain is {Py_GetAWSPortalRequestTypeEnumValue('Maintain')}")
                    if int(shutdownSetting) == int(Py_GetAWSPortalRequestTypeEnumValue("AutoShutdown")):                       
                        if slaveInfo.SlaveStatus == Py_GetSlaveStatusEnumValue("Rendering"):
                            if verbose:
                                self.LogInfo( instanceId + " is rendering." )
                            self.UpdateStatus(slaveName, self.Rendering)
                        elif status != None:
                            if verbose:
                                self.LogInfo( instanceId + " has stopped rendering." )
                            if int(slaveInfo.SlaveStatus) != int(Py_GetSlaveStatusEnumValue("Rendering")) and not status.startswith(self.LastRenderingTimePrefix): 
                                self.UpdateStatus(slaveName, self.LastRenderingTimePrefix + str(now))
                        else:
                            if verbose:
                                self.LogInfo(f"Shutdown setting is Autoshutdown. SlaveStatus is {int(slaveInfo.SlaveStatus)} and not equal to {int(Py_GetSlaveStatusEnumValue('Rendering'))}. AWSPortalWorker status is {status}")
                            self.UpdateStatus(slaveName, self.LastRenderingTimePrefix + str(now))# write the current time to the status file. The Worker is running and needs to be shut down even if it's idle for the whole life

                    elif int(shutdownSetting) == int(Py_GetAWSPortalRequestTypeEnumValue("Maintain")):
                        #if AutoShutdown is disabled, make the instance as such.
                        if verbose:
                            self.LogInfo("Instance is set to Maintain")
                        self.UpdateStatus(slaveName, self.Disabled)

                    else:
                        self.LogInfo("Reading the fleet type failed, will retry on the next check")
                        if verbose:
                            self.LogInfo("Got " + str(shutdownSetting) + " from  the shutdown type check")
                    
                    status = self.GetStatus(slaveName)

                    if status == self.Rendering or status.startswith(self.LastRenderingTimePrefix) :
                        if verbose:
                            self.LogInfo("Instance '{}' is enabled for Auto Shutdown".format(instanceId))

                        #We have previously determined that the instance is from AWS Portal. 
                        if int(slaveInfo.SlaveStatus) == int(Py_GetSlaveStatusEnumValue("Rendering")):
                            self.UpdateStatus(slaveName, self.Rendering)
                            if verbose:
                                self.LogInfo( instanceId + " is rendering." )
                        else:
                            if status != self.Rendering:
                                now = datetime.datetime.utcnow()
                                LastRenderingTime = self.GetLastRenderingTime(status)

                                if verbose:
                                    self.LogInfo( "Time Idle: " + str(now - LastRenderingTime) )
                                    
                                if now - LastRenderingTime > datetime.timedelta(minutes=self.GetIdleTime(instanceId)):
                                    if verbose:
                                        self.LogInfo( "Shutting Down: " + instanceId )
                                    
                                    try:
                                        self.DeleteStatus(slaveName)
                                        ec2 = self.GetEc2( slaveInfo.AWSInfo["AWSRegion"] )
                                        ec2.terminate_instances(InstanceIds=[instanceId])
                                    except Exception as e:
                                        self.LogInfo(str(e))
                            else:
                                self.UpdateStatus(slaveName, self.LastRenderingTimePrefix + str(now))
                                if verbose:
                                    self.LogInfo( instanceId + " is not rendering." )
                    else:
                        if verbose:
                            self.LogInfo( "Instance '{}' is not set to Auto Shutdown.".format(instanceId) )

                    if int(slaveInfo.SlaveStatus) == int(Py_GetSlaveStatusEnumValue("Offline")):
                        self.DeleteStatus(slaveName)

            else:
                if verbose:
                    self.LogInfo( "Not an AWS Worker" )
        
            if verbose:
                self.LogInfo( "AWS Portal Plugin - On Worker Info Updated Complete" )

        except:
            traceback.print_exc()

    def OnSlaveStopped( self, slaveName ):
        # type: (str) -> None
        slaveInfo = Py_GetSlaveInfo( slaveName, False )

        if hasattr(slaveInfo, "AWSInfo"):
            if "AWSInstanceId" in slaveInfo.AWSInfo.Keys:
                instanceId = slaveInfo.AWSInfo["AWSInstanceId"]
                self.DeleteStatus(slaveName)

    def CheckAutoShutdown( self, slaveInfo, verbose ):
        # type: (SlaveInfo, bool) -> int
        result = Py_GetAWSPortalRequestTypeEnumValue("AutoShutdown")#replaced Maintain default with Autoshutdown jusbla

        if verbose:
            self.LogInfo( "Checking if this Instance is from AWS Portal." )
        try:
            awsPortalInstance = Py_GetAWSPortalInstance(slaveInfo.AWSInfo['AWSInstanceId'])

            if awsPortalInstance:
                if verbose:
                    self.LogInfo( "Checking if the Request this instance came from is set to Auto Terminate." )
                awsPortalRequest = Py_GetAWSPortalFleetRequest( awsPortalInstance.DashFleetRequest )
                if awsPortalRequest is not None:
                    result = awsPortalRequest.AWSPortalRequestType
                    self.LogInfo(f"AWSPortalRequestType is {str(result)}")#jusbla

                    if result == Py_GetAWSPortalRequestTypeEnumValue("AutoShutdown") and verbose:
                        self.LogInfo( f"Instance is enabled for Auto Shutdown with result {result}" )#jusbla
                    elif verbose:
                        self.LogInfo( f"Instance is not enabled for Auto Shutdown with result {result}")#jusbla
                else:
                    self.LogInfo("Portal Fleet Request for our instance returned None")#jusbla

        except Exception as e:
            self.LogInfo("Error Checking Auto Shutdown.")
            self.LogInfo(str(e))
            

        return result

    def AssignPoolsAndGroups( self, slaveName, instanceId ):
        # type: (str, str) -> None
        self.WritePoolGroupCheckTime(datetime.datetime.utcnow())

        awsPortalInstance = Py_GetAWSPortalInstance( instanceId )
        if awsPortalInstance:
            awsPortalRequest = Py_GetAWSPortalFleetRequest( awsPortalInstance.DashFleetRequest )
            
            groupsAndPools = Py_GetEventPluginConfigMetaDataDictionary("AWSPortal")[awsPortalRequest.RequestId]

            # First try using the metadata, if that fails try querying the request object directly, if that fails throw an error.
            # Current versions of Deadline save the Pool and Group information in the EventPlugin Meta Data.
            # Previous versions had an issue where mismatched Deadline versions would overwrite the Group/Pool info that was in the Request objects, so the meta data is a way around that. 
            # We fall back to the previous way, we it doesn't work. (again, due to mismatches with clients and databases)
            if groupsAndPools is not None:
                groups, pools = self.DeserializeGroupsAndPools(groupsAndPools)

                self.SetGroupsAndPools(slaveName, groups, pools)
            else:
                groups = awsPortalRequest.Groups
                pools = awsPortalRequest.Pools

                if groups is not None and pools is not None:
                    self.SetGroupsAndPools(slaveName, groups, pools)
                else:
                    self.LogWarning("Could not assign groups or pools for fleet id {0}".format(awsPortalRequest.RequestId))
            
    def WritePoolGroupCheckTime(self, dateTime):
        # type: (datetime.datetime) -> None
        """Writes the last time the Pools and Groups for the Worker were checked."""
        try:
            PoolGroupCheck.SetLastPoolGroupCheckinTime(dateTime)
        except Exception as e:
            self.LogInfo(str(e))

    def GetIdleTime(self, instanceId):
        # type: (str) -> int
        awsPortalInstance = Py_GetAWSPortalInstance( instanceId )

        awsPortalRequest = Py_GetAWSPortalFleetRequest(awsPortalInstance.DashFleetRequest)

        return awsPortalRequest.IdleTime

    def GetEc2( self, region ):
        return boto3.client( 'ec2', region )

##############################################################################################
# These are C# functions we are wrapping in python so we can mock them for testing purposes. #
##############################################################################################
def Py_GetAWSPortalInstance(instanceId):
    # type: (str) -> DashFleetInstance
    return RepositoryUtils.GetAWSPortalInstance(instanceId)

def Py_GetAWSPortalFleetRequest(dash_fleet_request):
    # type: (DashFleetRequest) -> DashFleetRequest
    return RepositoryUtils.GetAWSPortalFleetRequest(dash_fleet_request)

def Py_GetAWSPortalRequestTypeEnumValue(enum_type):
    # type: (str) -> int
    return int(RepositoryUtils.GetAWSPortalRequestTypeEnumValue(enum_type))

def Py_GetEventPluginConfigMetaDataDictionary(meta_type):
    # type: (str) -> Dict[str, str]
    return RepositoryUtils.GetEventPluginConfigMetaDataDictionary(meta_type)

def Py_GetSlaveStatusEnumValue(enum_type):
    # type: (str) -> int
    return int(RepositoryUtils.GetSlaveStatusEnumValue(enum_type))

def Py_GetSlaveInfo(slaveName, invalidateCache):
    # type: (str, bool) -> SlaveInfo
    return RepositoryUtils.GetSlaveInfo(slaveName, invalidateCache)

def Py_GetSlaveSettings( slaveName, invalidateCache ):
    # type: (str, bool) -> SlaveSettings
    return RepositoryUtils.GetSlaveSettings(slaveName, invalidateCache)

def Py_SaveSlaveSettings(slaveSettings):
    # type: (SlaveSettings) -> None
    RepositoryUtils.SaveSlaveSettings(slaveSettings)

def Py_GetCurrentUserHomeDirectory():
    # type: () -> str
    return str(ClientUtils.GetCurrentUserHomeDirectory())
